/****************************************************************************************
 *
 * FileName:   MIDI_comms_lib.h 
 *
 * MIDI function library for AVR micro-controller  (AVR GCC under Atmel Studio)
 * =======================================================================================
 */
#ifndef _MIDI_COMMS_LIB_H
#define _MIDI_COMMS_LIB_H

#include <avr/io.h>
#include <stdlib.h>
#include "Pro_Micro_periph_lib.h"  // for low-level AVR UART Tx/Rx functions

#define UART1_putch(b)       UART_PutByte(b)     // function alias

#define OMNI_ON_POLY      1   // MIDI device responds in Poly mode on all channels
#define OMNI_ON_MONO      2   // MIDI device responds in Mono mode on all channels
#define OMNI_OFF_POLY     3   // MIDI device responds in Poly mode on base channel only
#define OMNI_OFF_MONO     4   // MIDI device responds in Mono mode on base channel only

#define NOTE_OFF_CMD         0x80    // 3-byte message
#define NOTE_ON_CMD          0x90    // 3-byte message
#define POLY_KEY_PRESS_CMD   0xA0    // 3-byte message
#define CONTROL_CHANGE_CMD   0xB0    // 3-byte message
#define PROGRAM_CHANGE_CMD   0xC0    // 2-byte message
#define CHAN_PRESSURE_CMD    0xD0    // 2-byte message
#define PITCH_BEND_CMD       0xE0    // 3-byte message
#define SYS_EXCLUSIVE_MSG    0xF0    // variable length message
#define SYSTEM_MSG_EOX       0xF7    // system msg terminator
#define SYS_EXCL_REMI_ID     0x73    // arbitrary pick... hope it's free!
#define CC_MODULATION        1       // Control change High byte
#define CC_BREATH_PRESSURE   2       //    ..     ..     ..
#define CC_CHANNEL_VOLUME    7       //    ..     ..     ..
#define CC_EXPRESSION        11      //    ..     ..     ..
#define MIDI_MSG_MAX_LENGTH  16      // not in MIDI specification!


// MIDI Channel Voice Messages ------------------------------------------------
/*
 * Function:     Transmit MIDI Note-On/Velocity message.
 *
 * Entry args:   chan = MIDI channel number (1..16)
 *
 *               noteNum = MIDI standard note number. (Note #60 = C4 = middle-C.)
 *
 *               velocity = initial output level (gain) of output envelope shaper,
 *                          typically, but depends on external MIDI device operation.
 *                          A value of zero should terminate the note.
 *
 * Output amplitude may be subsequently modified by a MIDI "after-touch" command,
 * or a Control Change message (e.g. breath pressure, expression, channel volume, etc),
 * if the connected MIDI device supports dynamic amplitude control.
 */
void   MIDI_SendNoteOn(uint8 chan, uint8 noteNum, uint8 velocity);

/*
 * Function:     Transmit MIDI Note-Off message.
 *
 * Entry args:   chan = MIDI channel number (1..16)
 *               noteNum = MIDI note number.
 *
 * Attn:         Note-Off Velocity is set to zero (0). Most MIDI devices ignore it.
 */
void   MIDI_SendNoteOff(uint8 chan, uint8 noteNum);

/*
 * Function:     Transmit MIDI Channel After-Touch (Pressure) message.
 *
 * Entry args:   chan = MIDI channel number (1..16)
 *               level = Aftertouch (pressure) level.
 */
void   MIDI_SendAfterTouch(uint8 chan, uint8 level);

/*
 * Function:     Transmit MIDI Pitch Bend message.
 *
 * Entry args:   chan = MIDI channel number (1..16)
 *               value = Pitch deviation value (14 bits). Unit depends on MIDI device.
 */
void   MIDI_SendPitchBend(uint8 chan, unsigned value);

/*
 * Function:     Transmit MIDI Control Change message.
 *
 * Entry args:   chan = MIDI channel number (1..16)
 *               ctrlNum = Control Number (0..119) -- not range checked.
 *               value = Controller data value (MSB or LSB).
 *
 * Examples:     ctrlNum = 0x01:0x21  Modulation lever, wheel, etc
 *               ctrlNum = 0x02:0x22  Breath controller
 *               ctrlNum = 0x07:0x27  Channel volume
 *               ctrlNum = 0x0B:0x2B  Expression controller
 *               ctrlNum = 0x0C:0x2C  Effects controller #1
 *               ctrlNum = 0x0D:0x2D  Effects controller #2
 *
 * The sound parameter affected by a Control Change message is dependent on the MIDI
 * sound module design and the patch (program) selected. It may be that some or all
 * controllers are unsupported, hence will have no effect on the sound generated.
 *
 * If ctrlNum < 0x20, value is controller MSB; else value is controller LSB.
 * If ctrlNum >= 0x40, controller data is single byte only.
 */
void   MIDI_SendControlChange(uint8 chan, uint8 ctrlNum, uint8 value);

/*
 * Function:     Transmit MIDI Program Change message.
 *
 * Entry args:   chan = MIDI channel number (1..16)
 *               progNum = Program (instrument/voice) number. Depends on MIDI device.
 */
void   MIDI_SendProgramChange(uint8 chan, uint8 progNum);


// MIDI Channel Mode and System Messages --------------------------------------
/*
 * Function:     Transmit MIDI Channel Omni/Poly/Mono (mode change) messages
 *               to set receiver mode. In Mono mode, only one channel is selected (M = 1).
 *
 * Entry args:   chan = MIDI channel number (1..16)
 *               mode = Omni/Poly/Mono  (mode = 1, 2, 3 or 4)
 *
 * Note:         This function also causes 'All Notes Off' to happen.
 */
void   MIDI_SendReceiverMode(uint8 chan, uint8 mode);

/*
 * Function:     Transmit MIDI Channel 'All Sound Off' (mode change) message.
 *               Status byte for Mode Change is the same as Control Change (0xBn).
 *
 * Entry args:   chan = MIDI channel number (1..16) = n + 1
 */
void   MIDI_SendAllSoundOff(uint8 chan);

/*
 * Function:     Transmit MIDI Channel 'Reset All Controllers' (mode change) message.
 *
 * Entry args:   chan = MIDI channel number (1..16)
 */
void   MIDI_SendResetAllControllers(uint8 chan);

// MIDI utility functions -----------------------------------------------------
/*
 * Function:     Find length of a given MIDI command/status message.
 *
 * Entry args:   MIDI command/status byte
 * 
 * Return:       (int) message length (bytes), usually 2 or 3, but...
 *               if status byte is SYS_EXCLUSIVE_CMD, return maximum length, or...
 *               if status byte is not supported (unrecognised), return 0.
 * 
 * todo:         support All Sound off, System reset, etc
 */
int    MIDI_GetMessageLength(uint8 statusByte);

#endif // _MIDI_COMMS_LIB_H
