/*
 * File:      midi_tester_main.c
 * `````
 * Project:  "MIDI Device Tester" 
 * ````````
 * Author:    M.J.Bauer  [ref. www.mjbauer.biz]
 * ```````
 * Hardware Platform:  "MIDI Rifficator" board using 'Pro Micro' MCU (ATmega32U4)
 * ```````````````````
 * Revision  History:  
 * ``````````````````````````````````````````````````````````````````````````````````````
 *     v1.0 | 19-AUG-2024 | Originated -- MJB
 *     v1.1 | 24-AUG-2024 | Fixed bug in ADC function in Pro_Micro_periph_lib
 *     v1.2 | 09-SEP-2024 | Fixed bug in legato note change
 *     v1.3 | 10-SEP-2024 | Added UI screen to send Control Change message
 *     v2.0 | 10-SEP-2024 | Added UI screen to send Program Change message
 *     v2.1 | 29-SEP-2024 | Added UI screen to set MIDI OUT channel (1..16)
 * 
 */ 
#include <avr/io.h>
#include "midi_tester_defs.h"

// ------------  Global variables and data  ---------------------------------------------
//
unsigned long  TX_LED_period_start;
unsigned long  millisLast;
unsigned task1Freq;        // Call frequency of TouchService() task (Hz)

uint8  midi_chan;          // MIDI channel # (1..16)
uint8  exprnCCnum;         // MIDI expression CC number (typ. 2, 7 or 11)
uint8  referenceNote;      // MIDI note number (for centre key)
uint8  notePlaying;        // MIDI note number (0 => No note playing)
uint8  notePlayingLast;    // MIDI note number (playing before new note)
uint8  velocity;           // MIDI velocity value (simulated)
uint8  modulation;         // MIDI modulation value (simulated)
uint8  expression;         // MIDI expression value (simulated)
uint8  keypadState;        // Keypad state (pads touched)

//============================================================================

void  setup(void)
{
    RX_LED_CONFIGURE();  // "system error" LED
    RX_LED_TURN_ON();
    TX_LED_CONFIGURE();  // "heartbeat" LED (debug option)
    TX_LED_TURN_OFF();
    SETUP_SWITCH_INPUT_PINS();
    
    SET_BIT(DDRC, 6);    // PC6 drives LED1
    SET_BIT(DDRD, 7);    // PD7 drives LED2
    SET_BIT(DDRE, 6);    // PE6 drives LED3
    SET_BIT(DDRD, 3);    // PD3 is TXD1 output
    CLEAR_BIT(DDRD, 4);  // PD4/ADC8 is Pot input
    
    TC0_Setup_SystemTimer();
    GLOBAL_INT_ENABLE();   // for timers
    UART_init(31250);
    ADC_Setup(ADC_REF_AVCC, ADC_CLOCK_1MHz);  // to read pot (ADC8)

    I2C_MasterInit(400);   // F_SCL = 400 kHz
    if (I2C_MasterStart(SH1106_I2C_ADDRESS) != 0)
        { while (1) ;;; }  // fault -- hang!
    LED1_TURN_ON();
    if (I2C_MasterSend(SH1106_MESSAGETYPE_COMMAND) != 0)
        { while (1) ;;; }  // fault -- hang!
    LED2_TURN_ON();
    if (I2C_MasterSend(SH1106_DISPLAYOFF) != 0)
        { while (1) ;;; }  // fault -- hang!
    LED3_TURN_ON();
    I2C_MasterStop();

    LED1_TURN_OFF();  // indicate I2C init successful
    LED2_TURN_OFF();
    LED3_TURN_OFF();

    SH1106_Init();
    SH1106_SetContrast(20);
    while (milliseconds() < 100) ;;;  // delay 100ms for OLED init
        
    SH1106_Test_Pattern(); // OLED display test
    while (milliseconds() < 700) ;;;  // delay 500ms to view test pattern

    Disp_Init();
    Disp_ClearScreen();
    GoToNextScreen(0);   // 0 => INITIALIZING

    RX_LED_TURN_OFF();   // init successful
    
    midi_chan = 1;
    exprnCCnum = 2;      // breath pressure
    referenceNote = 60;  // C4 (middle C)
    velocity = 80;
    modulation = 0;
    expression = 80;
}
    

void  loop()
{  
    static unsigned  task1Count;
    static bool  touch_cal_done;
    
    if ((milliseconds() - TX_LED_period_start) >= 50)
        TX_LED_TURN_OFF();   // 50ms ON-time ended
        
    if (milliseconds() != millisLast)  // Do 1ms periodic task...
    {
        millisLast = milliseconds();
        TouchService();
        task1Count++;
    }        
    
    if (isTaskPending_5ms())  // Do 5ms periodic tasks...
    {
        KeypadScan();    // Look for keypad strikes
        PlayerTask();    // Play notes when keyed
        PotService();    // Update Pot position
    }        

    if (isTaskPending_50ms())  // Do 50ms periodic tasks...
    {
        ButtonScan(4);         // Look for button hits
        UserInterfaceTask();   // User interface
        MidiControlChange();   // Send MIDI CC messages
    }   

    if (isTaskPending_500ms())  // Do 500ms periodic task... 
    {
        // Comment out next line to disable annoying "heartbeat" LED!
//      TX_LED_TURN_ON();  // Start LED period
        TX_LED_period_start = milliseconds();
        task1Freq = task1Count * 2;
        task1Count = 0;
    }    
    
    if (!touch_cal_done && milliseconds() > 2000)  // once only after MCU reset
        { TouchCalibrate();  touch_cal_done = TRUE; } 
}


int  main(void)
{
    setup();
    
    while (TRUE)  loop();   // loop forever
}


/*
 * Called at periodic intervals of 5 milliseconds from the main loop,
 * the "Player Task" (function) looks for keypad touch-on and touch-off events and
 * initiates or terminates notes accordingly.
 */
void  PlayerTask()
{
    if (milliseconds() < 2020)  return;  // initializing -- wait
    
    switch (keypadState)
    {
        case KEYS_OFF_OFF_OFF:
        {
            if (KeyStruck(LEFT_KEY))  InitiateNote(GO_DOWN_3);
            if (KeyStruck(CENTRE_KEY))  InitiateNote(FIXED_REF);
            if (KeyStruck(RIGHT_KEY))  InitiateNote(GO_UP_3);
            break;
        }        
        case KEYS_OFF_OFF_ON:
        {
            if (KeyStruck(LEFT_KEY))  InitiateNote(GO_DOWN_6);
            if (KeyStruck(CENTRE_KEY))  InitiateNote(GO_DOWN_3);
            if (KeyReleased(RIGHT_KEY))  TerminateNote();
            break;
        }        
        case KEYS_OFF_ON_OFF:
        {
            if (KeyStruck(LEFT_KEY))  InitiateNote(GO_DOWN_3);  
            if (KeyReleased(CENTRE_KEY))  TerminateNote();
            if (KeyStruck(RIGHT_KEY))  InitiateNote(GO_UP_3);
            break;
        }        
        case KEYS_OFF_ON_ON:
        {
            if (KeyStruck(LEFT_KEY))  InitiateNote(GO_DOWN_3);  
            if (KeyReleased(CENTRE_KEY))  InitiateNote(GO_UP_3);
            if (KeyReleased(RIGHT_KEY))  InitiateNote(GO_DOWN_3);
            break;
        }        
        case KEYS_ON_OFF_OFF:
        {
            if (KeyReleased(LEFT_KEY))  TerminateNote();
            if (KeyStruck(CENTRE_KEY))  InitiateNote(GO_UP_3);
            if (KeyStruck(RIGHT_KEY))  InitiateNote(GO_UP_6);
            break;
        }
        case KEYS_ON_OFF_ON:
        {
            if (KeyReleased(LEFT_KEY))  InitiateNote(GO_UP_6);
            if (KeyStruck(CENTRE_KEY))  InitiateNote(FIXED_REF);
            if (KeyReleased(RIGHT_KEY))  InitiateNote(GO_DOWN_6);
            break;
        }   
        case KEYS_ON_ON_OFF:
        {
            if (KeyReleased(LEFT_KEY))  InitiateNote(GO_UP_3);
            if (KeyReleased(CENTRE_KEY))  InitiateNote(GO_DOWN_3);
            if (KeyStruck(RIGHT_KEY))  InitiateNote(GO_UP_3);
            break;
        }        
        case KEYS_ON_ON_ON:
        {
            if (KeyReleased(LEFT_KEY))  InitiateNote(GO_UP_3);
            if (KeyReleased(CENTRE_KEY))  InitiateNote(FIXED_REF);
            if (KeyReleased(RIGHT_KEY))  InitiateNote(GO_DOWN_3);
            break;
        }  
    } // end switch
    
    keypadState = 0;     // Update keypad state...
    if (KeyTouched(LEFT_KEY)) keypadState |= 4;
    if (KeyTouched(CENTRE_KEY)) keypadState |= 2;
    if (KeyTouched(RIGHT_KEY)) keypadState |= 1;

    // Indicate keypad state on LEDs
    if (keypadState & (1<<2)) LED3_TURN_ON();  else LED3_TURN_OFF();  // LEFT key
    if (keypadState & (1<<1)) LED2_TURN_ON();  else LED2_TURN_OFF();  // CENTRE key
    if (keypadState & (1<<0)) LED1_TURN_ON();  else LED1_TURN_OFF();  // RIGHT key
}    


void  InitiateNote(uint8 direction)
{
    notePlaying = NoteSequence(direction);  // Get new note
    MIDI_SendNoteOn(midi_chan, notePlaying, velocity);
    if (notePlayingLast) MIDI_SendNoteOff(midi_chan, notePlayingLast);
    notePlayingLast = notePlaying;
}
    
    
void  TerminateNote()
{  
    if (notePlaying) 
    {
        MIDI_SendNoteOff(midi_chan, notePlaying);
        notePlayingLast = notePlaying = 0;
    }    
}


/*
 * Note Sequence Algorithm...
 *
 * Function returns a MIDI note number based on a deterministic sequence.
 * The argument (direction) determines whether the note returned should be fixed
 * (offset from the base note), a lower note, higher note, or the same as the last.
 *
 * Entry arg:  direction = FIXED_REF, SAME_AGAIN, GOING_DOWN, GOING_UP
 */
uint8  NoteSequence(uint8 direction)
{
    static uint8  previousNote;  // Last played note (if non-zero)
    uint8  selectedNote = previousNote;  // default (same again)

    // One-time exception on reset... initialize previousNote = referenceNote
    if (previousNote == 0) previousNote = referenceNote;
    
    if (direction == FIXED_REF)  selectedNote = referenceNote;
    if (direction == GO_DOWN_3)  selectedNote = previousNote - 3;
    if (direction == GO_UP_3)  selectedNote = previousNote + 3;
    if (direction == GO_DOWN_6)  selectedNote = previousNote - 6;
    if (direction == GO_UP_6)  selectedNote = previousNote + 6;
    
    if (selectedNote < 24) selectedNote = 24;
    if (selectedNote > 96) selectedNote = 96;

    previousNote = selectedNote;  // Save for next call
    return  selectedNote;
}


/*
 * Midi Control Change task
 *
 * This routine sends MIDI Control Change messages at 50ms intervals whenever
 * a CC data value changes, i.e. modulation (CC01) or expression (CC02).
 */
void   MidiControlChange()
{
    static uint8  modulationLast, expressionLast;
    
    if (modulation != modulationLast)
    {
        MIDI_SendControlChange(midi_chan, 1, modulation);
        modulationLast = modulation;
    }   
    
    if (expression != expressionLast)
    {
        MIDI_SendControlChange(midi_chan, exprnCCnum, expression);
        expressionLast = expression;
    }  
}    


//==================  T O U C H   K E Y - P A D   F U N C T I O N S   ===================
//
uint8  touch_onoff_thres[4];  // touch-pad on/off threshold values
uint8  touch_reading[4];      // touch-pad ADC readings (0..255)
bool   pad_touched[4];        // Pad on/off status (raw, from Touch Service routine)
bool   pad_touched_last[4];   // Pad on/off status (de-glitched, from Keypad Scan fn)
bool   key_struck[4], key_released[4];  // Pad on/off transition flags
/*
 * Function:  KeypadScan()
 *
 * Overview:  Touch keypad scanning routine, called periodically at 5 ms intervals.
 *
 *            Touchpad states are monitored for "touch-ON" and "touch-OFF" events.
 *            A "touch-ON" event is a transition from pad not touched to pad touched.
 *            A "touch-OFF" event is a transition from pad touched to pad not touched.
 *            Touchpad states are "de-glitched" using a 25ms "settling window".
 */
void  KeypadScan()
{
    static uint8  pad_touch_on_count[4];   // number of glitch-free touch-on samples
    static uint8  pad_touch_off_count[4];  // number of glitch-free touch-off samples
    uint8  pad;
    
    for (pad = 0;  pad < 3;  pad++)
    {
        if (!pad_touched[pad]) pad_touch_on_count[pad] = 0;
        if (pad_touched[pad]) pad_touch_off_count[pad] = 0;
        
        if (!pad_touched[pad] && pad_touch_off_count[pad] < 50)
            pad_touch_off_count[pad]++;  // capped at 50

        if (pad_touched[pad] && pad_touch_on_count[pad] < 50)
            pad_touch_on_count[pad]++;   // capped at 50
 
        if (pad_touched[pad] && !pad_touched_last[pad] && pad_touch_on_count[pad] >= 5) 
        {
            key_struck[pad] = TRUE;
            pad_touched_last[pad] = TRUE;
        }            
        
        if (!pad_touched[pad] && pad_touched_last[pad] && pad_touch_off_count[pad] >= 5) 
        {
            key_released[pad] = TRUE;
            pad_touched_last[pad] = FALSE;
        }            
    }
}    

/*
 * Function:  KeyTouched(pad)
 *
 * Overview:  Returns TRUE if the given key (pad) is currently being touched.
 *
 * Entry arg:   pad = digit identifying the pad to be checked...
 *                    0: Left pad, 1: Centre pad, 2: Right pad, 3: unused pad
 */
bool  KeyTouched(uint8 pad)
{
    return pad_touched_last[pad];
}
    
/*
 * Function:  KeyStruck(pad)
 *
 * Overview:  Returns TRUE if a "touch-ON event" has occurred since the previous call.
 *
 * Entry arg:   pad = digit identifying the pad to be checked...
 *                    0: Left pad, 1: Centre pad, 2: Right pad, 3: unused pad
 */
bool  KeyStruck(uint8 pad)
{
    bool result = FALSE;
    if (key_struck[pad]) result = TRUE;
    key_struck[pad] = FALSE;
    return result;
}
    
/*
 * Function:  KeyReleased(pad)
 *
 * Overview:  Returns TRUE if a "touch-OFF event" has occurred since the previous call.
 *
 * Entry arg:   pad = digit identifying the pad to be checked...
 *                    0: Left pad, 1: Centre pad, 2: Right pad, 3: unused pad
 */
bool  KeyReleased(uint8 pad)
{
    bool result = FALSE;
    if (key_released[pad]) result = TRUE;
    key_released[pad] = FALSE;
    return result;
}

/*
 * Function:  KeypadClear()
 *
 * Clear all key-struck and key-released flags to prevent false signaling of events.
 * To clear an individual flag, call respective function, e.g. KeyStruck(pad).
 */
void  KeypadClear()
{
    uint8  i;
    
    for (i = 0;  i < 4;  i++) 
    { 
        key_struck[i] = FALSE;  
        key_released[i] = FALSE; 
        pad_touched_last[i] = FALSE;
    }
}    

/*
 * Function:  TouchCalibrate()
 *
 * Determines touch-sense on/off threshold values which are written into array:
 * touch_onoff_thres[].
 * Function must be called at least once in the first second or two after MCU reset,
 * during which time the TouchService() routine must be executed periodically.
 * It is assumed the touch-pads are not being touched during the calibration period.
 */
void  TouchCalibrate()
{
    touch_onoff_thres[0] = touch_reading[0] - 10;
    touch_onoff_thres[1] = touch_reading[1] - 10;
    touch_onoff_thres[2] = touch_reading[2] - 10;
    touch_onoff_thres[3] = touch_reading[3] - 10;
    KeypadClear();
}    

/*
* TOUCH SENSE SERVICE ROUTINE
*
* Customized to read 4 touch inputs on PF4/ADC4, PF5/ADC5, PF6/ADC6 & PF7/ADC7.
* Current source is pin PB5.
*
* Called periodically at intervals of 1 millisec (approx).
* On each call, one touch input is serviced in sequence.
* Execution time is about 27 microseconds for each touch input.
* Global IRQ is disabled during execution of this function.
*/
void  TouchService()
{
    static  uint8  pad_idx;  // bit number (4..7) of Port F (ADC channel)
    uint8  pad;

    if (pad_idx == 0) pad_idx = 4; // begin with pad on PF4/ADC4
    
//  TESTPOINT_SET_HIGH();  // 'Scope trigger
    GLOBAL_INT_DISABLE();
   
    // Select ADC input for pad to be serviced (pad_idx) and enable ADC
    CLEAR_BIT(DDRF, pad_idx);  // configure pad pin as input
    ADCSRA = 0x03;             // Set prescaler to F_SYS/8 (ADC clock = 2 MHz)
    ADCSRB = 0x00;             // Set MUX5 bit = 0
    ADMUX = 0x60 + pad_idx;    // set ADC channel; Vref=AVCC; left-align result   
    SET_BIT(ADCSRA, ADEN);     // Enable ADC

    // Switch on the current source and delay for 6 microseconds (+/-1us)
    SET_BIT(PORTB, 5);
    SET_BIT(DDRB, 5);
    Delay_CPU_cycles(80);  // Macro defined in "pro-micro-periph.h"
    
    // Switch off the current source and read the ADC result
    CLEAR_BIT(PORTB, 5);
    CLEAR_BIT(DDRB, 5);
    SET_BIT(ADCSRA, ADSC);     // Start conversion (ADSC = 1)
    
    while (ADCSRA & (1<<ADSC))  { ; }  // wait until conversion done (ADSC == 0)
    
    pad = pad_idx - 4;         // pad ID number (0..3)
    touch_reading[pad] = ADCH; // get ADC result (MSB)
    SET_BIT(DDRF, pad_idx);    // Discharge pad -- configure pin as output (LOW)
    
    // Determine on/off state of touch-pad
    if (touch_reading[pad] < touch_onoff_thres[pad]) 
        pad_touched[pad] = TRUE; 
    else  pad_touched[pad] = FALSE; 

    if (++pad_idx == 8) pad_idx = 4;  // next pad
    
    GLOBAL_INT_ENABLE();
//  TESTPOINT_SET_LOW();
}    

// end-of-file
