/**
 * File:  midi_tester_ui.c
 *
 * MIDI Device Tester 'User Interface' comprising IIC OLED display, 4 push-buttons
 * and Data Entry potentiometer... (Rifficator H/W platform)
 *
 */
#include <avr/io.h>
#include "midi_tester_defs.h"

enum  User_Interface_Modes  // aka 'Screen identifiers'
{
    INITIALIZING = 0,
    HOME_SCREEN,
    SETTING_VELOCITY,
    SETTING_MODULATION,
    SETTING_EXPRESSION,
    SETTING_PITCH_BEND,
    SETTING_CHANNEL,
    SEND_PRGM_CHANGE,
    SEND_CTRL_CHANGE
};    

// ------------  Function prototype definitions  --------------------------------------
//
void   UserMode_Initializing();
void   UserMode_HomeScreen();
void   UserMode_SettingRefNote();
void   UserMode_SettingVelocity();
void   UserMode_SettingModulation();
void   UserMode_SettingExpression();
void   UserMode_SettingPitchBend();
void   UserMode_SettingChannel();
void   UserMode_SendProgramChange();
void   UserMode_SendControlChange();


//------------------- Private data ----------------------------------------------------
//
static bool   screenSwitchReq;       // flag: Request to switch to next screen
static bool   isNewScreen;           // flag: Screen switch has occurred
static uint8  currentScreen;         // ID number of current screen displayed
static uint8  previousScreen;        // ID number of previous screen displayed
static uint8  nextScreen;            // ID number of next screen to be displayed


/*
 * Display title bar text at top of screen in 12pt font with underline.
 * The title bar area is assumed to be already erased.
 * Maximum length of titleString is 16 ~ 17 chars.
 */
void  DisplayTitleBar(char *titleString)
{
    Disp_Mode(SET_PIXELS);
    Disp_PosXY(4, 0);
    Disp_SetFont(PROP_12_NORM);
    Disp_PutText(titleString);
    Disp_PosXY(0, 12);
    Disp_DrawLineHoriz(128);
}    


/*
 * Display a text string (4 chars max) centred in a button (bar) of fixed width (30 px)
 * using 8pt mono-spaced font, at the specified screen position (x, y) = bar upper LHS.
 * On exit, the display write mode is restored to 'SET_PIXELS'.
 */
void  DisplayButtonLegend(uint8 x, uint8 y, char *str)
{
    int  len = strlen(str);
    int  i;
    
    if (len > 4) len = 4;

    Disp_Mode(SET_PIXELS);  // Draw the button
    Disp_PosXY(x, y);
    Disp_DrawBar(30, 9);
    Disp_Mode(CLEAR_PIXELS);
    Disp_SetFont(MONO_8_NORM);  // Write the string
    x += 3 + 3 * (4 - len);  
    Disp_PosXY(x, y+1);
    
    for (i = 0;  i < len;  i++)  { Disp_PutChar(*str++); }
   
    Disp_Mode(SET_PIXELS);
}    


/*
 * Function returns the screen ID number of the currently displayed screen.
 */
uint8  GetCurrentScreenID()
{
    return  currentScreen;
}


/*
 * Function GoToNextScreen() triggers a screen switch to a specified screen ID.
 * The real screen switch business is done by the UserInterfaceTask() function
 * when next executed following the call to GoToNextScreen().
 *
 * Entry arg:  screenID = ID number of next screen required.
 */
void  GoToNextScreen(uint8 screenID)
{
    nextScreen = screenID;
    screenSwitchReq = TRUE;
}


/*
 * Called at periodic intervals of 50 milliseconds from the main loop, the
 * User Interface task allows the user to view and adjust various operational
 * parameters using the Data Entry Pot, 4 push-buttons and IIC OLED display.
 */
void  UserInterfaceTask(void)
{
    if (screenSwitchReq)   // Screen switch requested
    {
        screenSwitchReq = FALSE;
        isNewScreen = TRUE;  // Signal to render a new screen
        previousScreen = currentScreen;     // Make the switch...
        currentScreen = nextScreen;         // next screen => current
        if (nextScreen != previousScreen)  Disp_ClearScreen();
        PotMoved();  // clear 'pot moved' flag
    }
      
    switch (currentScreen)
    {
        case INITIALIZING:        UserMode_Initializing();       break;
        case HOME_SCREEN:         UserMode_HomeScreen();         break;
        case SETTING_VELOCITY:    UserMode_SettingVelocity();    break;
        case SETTING_MODULATION:  UserMode_SettingModulation();  break;
        case SETTING_EXPRESSION:  UserMode_SettingExpression();  break;
        case SETTING_PITCH_BEND:  UserMode_SettingPitchBend();   break;
        case SETTING_CHANNEL:     UserMode_SettingChannel();     break;
        case SEND_PRGM_CHANGE:    UserMode_SendProgramChange();  break;
        case SEND_CTRL_CHANGE:    UserMode_SendControlChange();  break;
        
    }
    
    isNewScreen = FALSE;
}


void  UserMode_Initializing()
{
    if (isNewScreen)
    {
        Disp_Mode(SET_PIXELS);
        Disp_SetFont(MONO_8_NORM);
        Disp_PosXY(3, 26);
        Disp_PutText("--- Initializing ---");
    }        
    
    // Exit this screen 2 sec. after reset
    if (milliseconds() > 2000)  GoToNextScreen(HOME_SCREEN);
}


void  UserMode_HomeScreen()
{
    static uint8  notePlayingLast;
    
    if (isNewScreen)
    {
        DisplayTitleBar("MIDI Tester  v2.1");
        Disp_SetFont(MONO_8_NORM);
        Disp_PosXY(8, 20);
        Disp_PutText("Note playing   Chan");
        Disp_SetFont(PROP_12_NORM);
        Disp_PosXY(104, 32);
        if (midi_chan < 10) Disp_PutChar(' ');
        Disp_PutDecimal(midi_chan, 1);
        DisplayButtonLegend(0, 53, "Noff");
        DisplayButtonLegend(32, 53, "PC");
        DisplayButtonLegend(64, 53, "CC");
        DisplayButtonLegend(96, 53, "SET");
    }       
    
    if (notePlaying != notePlayingLast)  // update note # display
    {
        Disp_SetFont(MONO_16_NORM);
        Disp_PosXY(36, 32);
        Disp_BlockClear(40, 16);  // clear existing data
        if (notePlaying == 0)  Disp_PutText("--");
        else  Disp_PutDecimal(notePlaying, 2);
        notePlayingLast = notePlaying;
    }        
    
    if (ButtonHit('A'))  MIDI_SendControlChange(midi_chan, 123, 0);  // All Notes Off
    if (ButtonHit('B'))  GoToNextScreen(SEND_PRGM_CHANGE);
    if (ButtonHit('C'))  GoToNextScreen(SEND_CTRL_CHANGE);
    if (ButtonHit('D'))  GoToNextScreen(SETTING_VELOCITY); 
}


void  UserMode_SettingVelocity()
{
    uint8  setting;
    bool  doRefresh = FALSE;
    
    if (isNewScreen)
    {
        DisplayTitleBar("Velocity");
        DisplayButtonLegend(0, 53, "Home");
        DisplayButtonLegend(32, 53, "Back");
        DisplayButtonLegend(64, 53, "x");
        DisplayButtonLegend(96, 53, "Next");
        doRefresh = TRUE;
    }
    
    if (PotMoved())
    {
        setting = (PotPosition() / 10) * 5;  // 0..125 step 5
        velocity = setting;
        doRefresh = TRUE;
    }        
    
    if (ButtonHit('A'))  GoToNextScreen(HOME_SCREEN);
    if (ButtonHit('B'))  GoToNextScreen(SETTING_PITCH_BEND);
    if (ButtonHit('C'))  DoNothing(); 
    if (ButtonHit('D'))  GoToNextScreen(SETTING_MODULATION);

    if (doRefresh)
    {
        Disp_SetFont(MONO_16_NORM);
        Disp_PosXY(48, 24);
        Disp_BlockClear(40, 16);  // clear existing data
        Disp_PutDecimal(velocity, 2);
    }        
}   


void  UserMode_SettingModulation()
{
    uint8  setting;
    bool  doRefresh = FALSE;
    
    if (isNewScreen)
    {
        DisplayTitleBar("Modulation");
        DisplayButtonLegend(0, 53, "Home");
        DisplayButtonLegend(32, 53, "Back");
        DisplayButtonLegend(64, 53, "x");
        DisplayButtonLegend(96, 53, "Next");
        doRefresh = TRUE;
    }
    
    if (PotMoved())
    {
        setting = (PotPosition() / 10) * 5;  // 0..125 step 5
        modulation = setting;
        doRefresh = TRUE;
    }        
    
    if (ButtonHit('A'))  GoToNextScreen(HOME_SCREEN);
    if (ButtonHit('B'))  GoToNextScreen(SETTING_VELOCITY);
    if (ButtonHit('C'))  DoNothing(); 
    if (ButtonHit('D'))  GoToNextScreen(SETTING_EXPRESSION);

    if (doRefresh)
    {
        Disp_SetFont(MONO_16_NORM);
        Disp_PosXY(48, 24);
        Disp_BlockClear(40, 16);  // clear existing data
        Disp_PutDecimal(modulation, 2);
    }      
}   


void   UserMode_SettingExpression()
{
    static uint8  lastExprnCCnum;
    uint8  setting;
    bool  doRefresh = FALSE;
    
    if (isNewScreen)
    {
        DisplayTitleBar("Expression");
        DisplayButtonLegend(0, 53, "Home");
        DisplayButtonLegend(32, 53, "Back");
        DisplayButtonLegend(64, 53, "CC#");
        DisplayButtonLegend(96, 53, "Next");
        lastExprnCCnum = 99;  // force refresh
        doRefresh = TRUE;
    }
    
    if (PotMoved())
    {
        setting = (PotPosition() / 10) * 5;  // 0..125 step 5
        expression = setting;
        doRefresh = TRUE;
    }        
    
    if (ButtonHit('A'))  GoToNextScreen(HOME_SCREEN);
    if (ButtonHit('B'))  GoToNextScreen(SETTING_MODULATION);
    if (ButtonHit('C'))  // Scroll CC number for expres'n
    {
        if (exprnCCnum == 2) exprnCCnum = 7;
        else if (exprnCCnum == 7) exprnCCnum = 11;
        else  exprnCCnum = 2;
    }        
    if (ButtonHit('D'))  GoToNextScreen(SETTING_PITCH_BEND);

    if (doRefresh)
    {
        Disp_SetFont(MONO_16_NORM);
        Disp_PosXY(48, 22);
        Disp_BlockClear(40, 16);  // clear existing data
        Disp_PutDecimal(expression, 2);
    }
    if (exprnCCnum != lastExprnCCnum)
    { 
        Disp_SetFont(PROP_8_NORM);
        Disp_PosXY(8, 42);
        Disp_BlockClear(112, 8);  // clear existing data
        if (exprnCCnum == 2)  Disp_PutText("CC2 Breath Controller");
        if (exprnCCnum == 7)  Disp_PutText("CC7 Channel Volume");
        if (exprnCCnum == 11)  Disp_PutText("CC11 Expression Pedal");
        lastExprnCCnum = exprnCCnum;
    }      
}    


void   UserMode_SettingPitchBend()
{
    static int  pitchBend_14b = 0x2000;  // centre pos
    int  pitchBend_PC = 0;
    bool  doRefresh = FALSE;
    
    if (isNewScreen)
    {
        DisplayTitleBar("Pitch Bend");
        DisplayButtonLegend(0, 53, "Home");
        DisplayButtonLegend(32, 53, "Back");
        DisplayButtonLegend(64, 53, "Zero");
        DisplayButtonLegend(96, 53, "Next");
        doRefresh = TRUE;
    }
    
    if (PotMoved())
    {
        pitchBend_14b = (int) PotPosition() << 6;  // 0..16K
        MIDI_SendPitchBend(midi_chan, pitchBend_14b);
        doRefresh = TRUE;
    }        
    
    if (ButtonHit('A'))  
    {
        pitchBend_14b = 0x2000;  // centre pos
        MIDI_SendPitchBend(midi_chan, pitchBend_14b);
        GoToNextScreen(HOME_SCREEN);
    }        
    if (ButtonHit('B'))  GoToNextScreen(SETTING_EXPRESSION);
    if (ButtonHit('C'))  // Zero the control setting
    {
        pitchBend_14b = 0x2000;  // centre pos
        MIDI_SendPitchBend(midi_chan, pitchBend_14b);
        doRefresh = TRUE;
    }        
    if (ButtonHit('D'))  GoToNextScreen(SETTING_CHANNEL);

    if (doRefresh)
    {
        pitchBend_PC = ((long)(pitchBend_14b - 0x1FFF) * 100) / 0x2000;  // percent FS
        Disp_SetFont(MONO_16_NORM);
        Disp_PosXY(40, 24);
        Disp_BlockClear(40, 16);  // clear existing data
        if (pitchBend_PC < 0)  Disp_PutChar('-');  
        else  Disp_PutChar('+');
        Disp_PutDecimal(abs(pitchBend_PC), 2);  // Display as 0..+/-100 %
        Disp_SetFont(PROP_12_NORM);
        Disp_PutText(" %");
    }      
}    


void  UserMode_SettingChannel()
{
    bool   doRefresh = FALSE;
    
    if (isNewScreen)
    {
        DisplayTitleBar("MIDI channel");
        DisplayButtonLegend(0, 53, "Home");
        DisplayButtonLegend(32, 53, "Back");
        DisplayButtonLegend(64, 53, "x");
        DisplayButtonLegend(96, 53, "Next");
        doRefresh = TRUE;
    }
    
    if (PotMoved())
    {
        midi_chan = PotPosition() / 16 + 1;  // 1..16
        doRefresh = TRUE;
    }
    
    if (ButtonHit('A'))  GoToNextScreen(HOME_SCREEN);
    if (ButtonHit('B'))  GoToNextScreen(SETTING_PITCH_BEND);
    if (ButtonHit('C'))  DoNothing();
    if (ButtonHit('D'))  GoToNextScreen(SETTING_VELOCITY);  // wrap

    if (doRefresh)
    {
        Disp_SetFont(MONO_16_NORM);
        Disp_PosXY(48, 24);
        Disp_BlockClear(40, 16);
        Disp_PutDecimal(midi_chan, 2);
    }
}


void  UserMode_SendProgramChange()
{
    static uint8  prgmRange, prgmRangeLast, prgmNumber;
    bool   doRefresh = FALSE;
    uint8  setting;
    
    if (isNewScreen)
    {
        DisplayTitleBar("Program Change");
        Disp_SetFont(MONO_8_NORM);
        Disp_PosXY(20, 42);
        Disp_PutText("Range: ");
        DisplayButtonLegend(0, 53, "Home");
        DisplayButtonLegend(32, 53, "Rng");
        DisplayButtonLegend(64, 53, "x");
        DisplayButtonLegend(96, 53, "Send");
        prgmRangeLast = 99;  // force refresh
        doRefresh = TRUE;
    }
    
    if (PotMoved())
    {
        setting = PotPosition() / 8;  // 0..31
        prgmNumber = (prgmRange * 32) + setting;
        doRefresh = TRUE;
    }    
    
    if (ButtonHit('A'))  GoToNextScreen(HOME_SCREEN);
    if (ButtonHit('B'))  prgmRange = (prgmRange + 1) % 4;  // 0..3
    if (ButtonHit('C'))  DoNothing();
    if (ButtonHit('D'))  MIDI_SendProgramChange(midi_chan, prgmNumber);

    if (doRefresh)
    {
        Disp_SetFont(MONO_16_NORM);
        Disp_PosXY(48, 20);
        Disp_BlockClear(40, 16);  // clear existing data
        Disp_PutDecimal(prgmNumber, 3);
    }
    if (prgmRange != prgmRangeLast)  // range changed
    { 
        Disp_SetFont(MONO_8_NORM);
        Disp_PosXY(64, 42);
        Disp_BlockClear(56, 8);  // clear existing data
        if (prgmRange == 0)  Disp_PutText("0..31");
        if (prgmRange == 1)  Disp_PutText("32..63");
        if (prgmRange == 2)  Disp_PutText("64..95");
        if (prgmRange == 3)  Disp_PutText("96..127");
        prgmRangeLast = prgmRange;
    }      
}    


void  UserMode_SendControlChange()
{
    static uint8  ccNumberList[] =  { 0, 3, 4, 5, 6, 16, 17, 18, 19, 64, 65, 66, 67, 68, 
                                     84, 85, 86, 100, 101, 102, 103, 104, 105, 106 };
    static uint8  ccNumber, ccData;
    static bool   potSetsCCnum = TRUE;  // one-time initialization at start-up
    bool   doRefresh = FALSE;
    uint8  ccListSize = sizeof(ccNumberList);
    uint8  setting;
    
    if (isNewScreen)
    {
        DisplayTitleBar("Control Change");
        Disp_SetFont(MONO_8_NORM);
        Disp_PosXY(24, 18);
        Disp_PutText("CC #");
        Disp_PosXY(80, 18);
        Disp_PutText("Data");
        DisplayButtonLegend(0, 53, "Home");
        DisplayButtonLegend(32, 53, "CC#");
        DisplayButtonLegend(64, 53, "Data");
        DisplayButtonLegend(96, 53, "Send");
        doRefresh = TRUE;
    }
    
    if (PotMoved())
    {
        if (potSetsCCnum)
        {
            setting = PotPosition() / 10;  // 0..25
            if (setting >= ccListSize) setting = ccListSize - 1;
            ccNumber = ccNumberList[setting];
        }
        else  // pot sets ccData
        {
            setting = (PotPosition() / 10) * 5;  // 0..125 step 5
            ccData = setting;
        }            
        doRefresh = TRUE;
    }    
    
    if (ButtonHit('A'))  GoToNextScreen(HOME_SCREEN);
    if (ButtonHit('B'))  
    {
        potSetsCCnum = TRUE;  // Setting CC number
        doRefresh = TRUE;
    }        
    if (ButtonHit('C'))  
    {
        potSetsCCnum = FALSE;  // Setting CC data value
        doRefresh = TRUE;
    }        
    if (ButtonHit('D'))  MIDI_SendControlChange(midi_chan, ccNumber, ccData);

    if (doRefresh)
    {
        Disp_SetFont(MONO_16_NORM);
        Disp_PosXY(24, 30);
        Disp_BlockClear(96, 20);  // clear existing data
        Disp_PutDecimal(ccNumber, 3);
        Disp_PosXY(80, 30);
        Disp_PutDecimal(ccData, 3);
        if (potSetsCCnum) Disp_PosXY(24, 46);  else Disp_PosXY(80, 46);
        Disp_DrawBar(34, 2);  // Underscore item for setting
    }         
}    


//===================   P U S H - B U T T O N   F U N C T I O N S  ======================
//
const char button_code[] = { 'A', 'B', 'C', 'D' };  // ASCII code table
static bool  is_button_hit[4];   // TRUE if button[x] hit detected

/*
 * Function ButtonScan() must be called periodically from the application program
 * (main loop) at intervals of about 30 ~ 50ms for reliable "de-bounce" operation.
 * Button inputs must have MCU internal pull-ups enabled if no external pull-ups.
 *
 * Entry argument 'nButts' specifies the number of buttons (1..4) to be serviced.
 * For example, if nButts is 1, only Button_A is serviced;  if nButts is 3, then 3
 * buttons (Button_A, Button_B and Button_C) will be serviced.
 */
void  ButtonScan(unsigned char nButts)
{
    static bool  button_down_last_scan[4];
    bool  is_button_down;   // TRUE if button is pressed
    unsigned char x;        // button array index (0..3)

    if (nButts > 4)  nButts = 4;

    for (x = 0 ; x < nButts ; x++)    // Scan buttons up to nButts
    {
        is_button_down = !TEST_BIT(PINB, (x + 1));  // Read button input state

        if (is_button_down && !button_down_last_scan[x])  // hit detected
            is_button_hit[x] = TRUE;
 
        button_down_last_scan[x] = is_button_down;  // update status
    }
}

/*
 * Function ButtonHit() returns the Boolean value (TRUE or FALSE) of a flag indicating
 * whether or not a "button hit" event occurred since the previous call to the function.
 *
 * Entry argument 'button_ID' is an ASCII code identifying one of 4 buttons to check,
 * which must be one of: 'A', 'B', 'C' or 'D', otherwise the function will return FALSE.
 */
bool  ButtonHit(char  button_ID)
{
    bool  result = FALSE;
    uint8  x;

    for (x = 0 ; x < 4 ; x++)
    {
        if (button_ID == button_code[x])   // Found button_ID
        {
            result = is_button_hit[x];
            is_button_hit[x] = FALSE;
            break;
        }
    }
    return result;
}


//----------------------------------------------------------------------------------------
// Functions to support data-entry potentiometer
//----------------------------------------------------------------------------------------
//
static unsigned long  potReadingAve;   // smoothed pot reading [24:8 fixed-pt]
//
/*
 * Function:  PotService()
 *
 * Overview:  Service Routine for front-panel data-entry pot.
 *            Non-blocking "task" called at 5ms intervals from main loop.
 *
 * Detail:    The routine reads the pot input and keeps a rolling average of the ADC
 *            readings in fixed-point format (24:8 bits);  range 0.0 ~ 1023.0
 *            The current pot position can be read by a call to function PotPosition().
 */
void  PotService()
{
    long  potReading = ADC_ReadInput(ADC_8);  // 10-bit reading;
    
    // Apply 1st-order IIR filter (K = 0.25)
    potReading = potReading << 8;  // convert to fixed-point (24:8 bits)
    potReadingAve -= (potReadingAve >> 2);
    potReadingAve += (potReading >> 2);
}

/*
 * Function:     PotMoved()
 *
 * Overview:     Returns TRUE if the pot position has changed by more than 2% since a
 *               previous call to the function which also returned TRUE.
 *
 * Return val:   (bool) status flag, value = TRUE or FALSE
 */
bool  PotMoved()
{
    static long  lastReading;
    bool  result = FALSE;

    if (labs(potReadingAve - lastReading) > (20 << 8))  // 20/1024 -> 2%
    {
        result = TRUE;
        lastReading = potReadingAve;  
    }
    return  result;
}

/*
 * Function:     PotPosition()
 *
 * Overview:     Returns the current position of the data-entry pot, averaged over 
 *               several ADC readings, as an 8-bit unsigned integer.
 *
 * Return val:   (uint8) Filtered Pot position reading, range 0..255.
 */
uint8  PotPosition()
{
    return  (uint8) (potReadingAve >> 10);  // = (Integer part) / 4
}

